#include "audiowidget.h"
#include "miniaudio.c"

AudioWidget *AudioWidget::_instance = 0;

AudioWidget *AudioWidget::getInstance() {

    if (_instance == 0) {
        _instance = new AudioWidget();
        Q_CHECK_PTR(_instance);
    }
    return _instance;
}

AudioWidget::AudioWidget() :
    engineCount(0)
{
    this->startEngine();
    layout = new QHBoxLayout();
    for (int i= 0; i < Settings::getInstance()->getLayersNumber(); i++ ) {
        layout->insertWidget(i, new AudioLayerWidget(this, tr("Layer %1").arg(i + 1)));
     }
    setLayout(layout);
}

void data_callback(ma_device* pDevice, void* pOutput, const void* pInput, ma_uint32 frameCount)
{
    (void)pInput;
    //Do master audio processing before sending to device.
    ma_engine_read_pcm_frames((ma_engine*)pDevice->pUserData, pOutput, frameCount, NULL);
}

void AudioWidget::startEngine(int n)
{
    this->startContext();
    this->startDevice(n);
}

void AudioWidget::startDevice(int id)
{
    ma_result result;
    ma_device_config deviceConfig;
    ma_engine_config engineConfig;

    deviceConfig = ma_device_config_init(ma_device_type_playback);
    deviceConfig.playback.pDeviceID = &pPlaybackDeviceInfos[id].id;
    deviceConfig.playback.format    = resourceManager.config.decodedFormat;
    deviceConfig.playback.channels  = 0;
    deviceConfig.sampleRate         = resourceManager.config.decodedSampleRate;
    deviceConfig.dataCallback       = data_callback;
    deviceConfig.pUserData          = &engines[engineCount];
    result = ma_device_init(&context, &deviceConfig, &devices[engineCount]);
    if (result != MA_SUCCESS) {
        qCritical("Failed to initialize audio device %s.", pPlaybackDeviceInfos[id].name);
        return;
    }
    engineConfig = ma_engine_config_init();
    engineConfig.pDevice          = &devices[engineCount];
    engineConfig.pResourceManager = &resourceManager;
    engineConfig.noAutoStart      = MA_TRUE;
    result = ma_engine_init(NULL, &engines[engineCount]);
    if (result != MA_SUCCESS) {
        qCritical("Failed to initialize audio engine.");
        return;
    }
    result = ma_engine_start(&engines[engineCount]);
    if (result != MA_SUCCESS) {
        qCritical("Failed to start audio engine %d.", engineCount);
    }
    //engineCount +=1;
    iChosenDevice = id;
    qInfo("Initialized audio device %d: %s", id, pPlaybackDeviceInfos[id].name);
}

void AudioWidget::startContext()
{
    ma_result result;

    resourceManagerConfig = ma_resource_manager_config_init();
    resourceManagerConfig.decodedFormat     = ma_format_f32;    /* ma_format_f32 should almost always be used as that's what the engine (and most everything else) uses for mixing. */
    resourceManagerConfig.decodedChannels   = 0;
    resourceManagerConfig.decodedSampleRate = 0;
    resourceManagerConfig.jobThreadCount    = 0;
    result = ma_resource_manager_init(&resourceManagerConfig, &resourceManager);
    if (result != MA_SUCCESS) {
        qCritical("Failed to initialize audio resource manager.");
        return;
    }
    result = ma_context_init(NULL, 0, NULL, &context);
    if (result != MA_SUCCESS) {
        qCritical("Failed to initialize audio context.");
        return;
    }
}

void AudioWidget::startEngine()
{
    this->startContext();
    this->getAllAudioDevices();
    iChosenDevice = Settings::getInstance()->getAudioDeviceId();
    this->startDevice(iChosenDevice);
}

ma_engine AudioWidget::getEngine()
{
    return(engines[engineCount]);
}

void AudioWidget::stopEngine()
{
    ma_engine_uninit(&engines[engineCount]);
    ma_device_uninit(&devices[engineCount]);
    ma_context_uninit(&context);
    ma_resource_manager_uninit(&resourceManager);
}

void AudioWidget::mediaLoaded(int layer, QString media)
{
    QLayoutItem * const item = layout->itemAt(layer);
    dynamic_cast<AudioLayerWidget *>(item->widget())->loadMedia(media);
}

void AudioWidget::volChanged(int layer, qreal vol) {
    QLayoutItem * const item = layout->itemAt(layer);
    dynamic_cast<AudioLayerWidget *>(item->widget())->setVol(vol);
}

void AudioWidget::panChanged(int layer, qreal vol) {
    QLayoutItem * const item = layout->itemAt(layer);
    dynamic_cast<AudioLayerWidget *>(item->widget())->setPan(vol);
}

void AudioWidget::pitchChanged(int layer, qreal vol) {
    QLayoutItem * const item = layout->itemAt(layer);
    dynamic_cast<AudioLayerWidget *>(item->widget())->setPitch(vol);
}

void AudioWidget::playbackChanged(int layer, Status status)
{
    QLayoutItem * const item = layout->itemAt(layer);
    switch (status) {
    case PlayingOnce:
      dynamic_cast<AudioLayerWidget *>(item->widget())->play(false);
      break;
    case Stopped:
      dynamic_cast<AudioLayerWidget *>(item->widget())->stop();
      break;
    case Paused:
      dynamic_cast<AudioLayerWidget *>(item->widget())->pause();
      break;
    case PlayingLoop:
      dynamic_cast<AudioLayerWidget *>(item->widget())->play(true);
      break;
    }
}

// enum all audio devices in system
void AudioWidget::getAllAudioDevices()
{
    ma_result result;

    result = ma_context_get_devices(&context, &pPlaybackDeviceInfos, &playbackDeviceCount, NULL, NULL);
    if (result != MA_SUCCESS) {
        qCritical("Failed to enumerate playback devices.");
        ma_context_uninit(&context);
        return;
    }
    for (ma_uint32 iAvailableDevice = 0; iAvailableDevice < playbackDeviceCount; iAvailableDevice += 1) {
        qInfo("%d: : %s", iAvailableDevice, pPlaybackDeviceInfos[iAvailableDevice].name);
    }
    return ;
}
